@@ -1,9 +1,13 @@
 from .api import *
+from .data import *
 from .models import *
 from .protocol import *
 from .inference import *
 from .configs import Configs
-from .data import PrefetchLoader
+from .protocol import PrefetchLoader
+from .misc._api import _remove
+from .misc._api import _rmtree
 from .modules.heads import HeadBase
 from .modules.heads import HeadConfigs
 from .modules.extractors import ExtractorBase
+from .modules.aggregators import AggregatorBase
