@@ -22,30 +22,28 @@ __all__ = [
     "Auto",
     # basic
     "make",
+    "make_from",
+    "finetune",
     "save",
     "load",
     "evaluate",
-    "load_task",
+    "task_loader",
+    "load_experiment_results",
     "repeat_with",
-    "tasks_to_pipelines",
-    "tasks_to_patterns",
-    "transform_experiments",
     "make_toy_model",
     "Task",
-    "Experiments",
+    "Experiment",
     "ModelPattern",
     "EnsemblePattern",
     "RepeatResult",
-    "SAVING_DELIM",
-    "_remove",
-    "_rmtree",
     # ensemble
     "Ensemble",
     "EnsembleResults",
     # register
-    "register_pipe",
-    "register_head",
     "register_extractor",
+    "register_head",
+    "register_aggregator",
+    "register_pipe",
     "register_model",
     "register_config",
     "register_head_config",
@@ -61,4 +59,6 @@ __all__ = [
     "PipeInfo",
     # production
     "Pack",
+    "PackModel",
+    "PipelineModel",
 ]
