@@ -1,13 +1,14 @@
 import os
-import time
-import shutil
+
+import numpy as np
 
 from typing import *
 from tqdm.autonotebook import tqdm
 from cftool.misc import update_dict
 from cftool.misc import shallow_copy_dict
+from cftool.misc import lock_manager
+from cftool.misc import Saving
 from cftool.misc import LoggingMixin
-from cftool.ml.utils import pattern_type
 from cftool.ml.utils import patterns_type
 from cftool.ml.utils import Comparer
 from cftool.ml.utils import Estimator
@@ -18,21 +19,112 @@ from cfdata.tabular import parse_task_type
 from cfdata.tabular import TaskTypes
 
 from ..dist import Task
-from ..dist import Experiments
+from ..dist import Experiment
+from ..dist import ExperimentResults
 from ..types import data_type
+from ..types import general_config_type
+from ..configs import _parse_config
 from ..trainer import Trainer
 from ..trainer import IntermediateResults
 from ..pipeline import Pipeline
 from ..protocol import DataProtocol
+from ..misc._api import _to_saving_path
+from ..misc._api import _make_saving_path
+from ..misc._api import _fetch_saving_paths
 from ..misc.toolkit import to_2d
 
 
-def make(model: str = "fcnn", **kwargs: Any) -> Pipeline:
+def make(
+    model: str = "fcnn",
+    config: general_config_type = None,
+    increment_config: general_config_type = None,
+    **kwargs: Any,
+) -> Pipeline:
     kwargs["model"] = model
-    return Pipeline.make(kwargs)
+    config, increment_config = map(_parse_config, [config, increment_config])
+    return Pipeline.make(update_dict(increment_config, update_dict(config, kwargs)))
+
+
+def make_from(
+    identifier: str = "cflearn",
+    saving_folder: Optional[str] = None,
+    kwargs_callback: Optional[Callable[[Dict[str, Any]], None]] = None,
+    cuda: Optional[Union[int, str]] = None,
+    *,
+    model_mapping: Optional[Dict[str, str]] = None,
+) -> Dict[str, List[Pipeline]]:
+    saving_folder = os.path.abspath(saving_folder or "./")
+
+    def _core() -> Pipeline:
+        compress = os.path.isfile(f"{path}.zip")
+        with lock_manager(saving_folder, [path]):
+            with Saving.compress_loader(path, compress):
+                kwargs = Saving.load_dict("config", path)
+        if model_mapping is not None:
+            kwargs["model"] = model_mapping[model]
+        if cuda is not None:
+            kwargs["cuda"] = cuda
+        kwargs.pop("logging_folder", None)
+        if kwargs.get("mlflow_config") is not None:
+            kwargs["mlflow_config"] = {}
+        if kwargs_callback is not None:
+            kwargs_callback(kwargs)
+        return make(**kwargs)
+
+    ms: Dict[str, List[Pipeline]] = {}
+    paths_dict = _fetch_saving_paths(identifier, saving_folder)
+    for model, paths in paths_dict.items():
+        for path in paths:
+            ms.setdefault(model, []).append(_core())
+    return ms
+
+
+def finetune(
+    x: data_type,
+    y: data_type = None,
+    x_cv: data_type = None,
+    y_cv: data_type = None,
+    *,
+    model: Optional[str] = None,
+    strict: bool = True,
+    identifier: str = "cflearn",
+    pretrain_folder: Optional[str] = None,
+    kwargs_callback: Optional[Callable[[Dict[str, Any]], None]] = None,
+    state_dict_callback: Optional[Callable[[Dict[str, Any]], None]] = None,
+    sample_weights: Optional[np.ndarray] = None,
+    cuda: Optional[Union[int, str]] = None,
+) -> Pipeline:
+    paths_dict = _fetch_saving_paths(identifier, pretrain_folder)
+    all_paths: List[str] = sum(paths_dict.values(), [])
+    if len(all_paths) > 1:
+        raise ValueError("more than 1 model is detected")
+    if model is None:
+        model_mapping = None
+    else:
+        pretrain_model = list(paths_dict.keys())[0]
+        model_mapping = {pretrain_model: model}
+    ms = make_from(
+        identifier,
+        pretrain_folder,
+        kwargs_callback,
+        cuda,
+        model_mapping=model_mapping,
+    )
+    m = list(ms.values())[0][0]
+    m.fit(
+        x,
+        y,
+        x_cv,
+        y_cv,
+        pretrain_strict=strict,
+        pretrain_folder=pretrain_folder,
+        pretrain_identifier=identifier,
+        state_dict_callback=state_dict_callback,
+        sample_weights=sample_weights,
+    )
+    return m
 
 
-SAVING_DELIM = "^_^"
 pipelines_type = Union[
     Pipeline,
     List[Pipeline],
@@ -41,36 +133,6 @@ pipelines_type = Union[
 ]
 
 
-def _to_saving_path(identifier: str, saving_folder: Optional[str]) -> str:
-    if saving_folder is None:
-        saving_path = identifier
-    else:
-        saving_path = os.path.join(saving_folder, identifier)
-    return saving_path
-
-
-def _make_saving_path(
-    i: int,
-    name: str,
-    saving_path: str,
-    remove_existing: bool,
-) -> str:
-    saving_path = os.path.abspath(saving_path)
-    saving_folder, identifier = os.path.split(saving_path)
-    postfix = f"{SAVING_DELIM}{name}{SAVING_DELIM}{i:04d}"
-    if os.path.isdir(saving_folder) and remove_existing:
-        for existing_model in os.listdir(saving_folder):
-            if os.path.isdir(os.path.join(saving_folder, existing_model)):
-                continue
-            if existing_model.startswith(f"{identifier}{postfix}"):
-                print(
-                    f"{LoggingMixin.warning_prefix}"
-                    f"'{existing_model}' was found, it will be removed"
-                )
-                os.remove(os.path.join(saving_folder, existing_model))
-    return f"{saving_path}{postfix}"
-
-
 def _to_pipelines(pipelines: pipelines_type) -> Dict[str, List[Pipeline]]:
     if isinstance(pipelines, dict):
         pipeline_dict = {}
@@ -182,7 +244,7 @@ def save(
     identifier: str = "cflearn",
     saving_folder: Optional[str] = None,
     *,
-    retain_data: bool = True,
+    compress: bool = True,
 ) -> Dict[str, List[Pipeline]]:
     pipeline_dict = _to_pipelines(pipelines)
     saving_path = _to_saving_path(identifier, saving_folder)
@@ -190,42 +252,20 @@ def save(
         for i, pipeline in enumerate(pipeline_list):
             pipeline.save(
                 _make_saving_path(i, name, saving_path, True),
-                retain_data=retain_data,
-                compress=True,
+                compress=compress,
             )
     return pipeline_dict
 
 
-def _fetch_saving_paths(
-    identifier: str = "cflearn",
-    saving_folder: Optional[str] = None,
-) -> Dict[str, List[str]]:
-    paths: Dict[str, List[str]] = {}
-    saving_path = _to_saving_path(identifier, saving_folder)
-    saving_path = os.path.abspath(saving_path)
-    base_folder = os.path.dirname(saving_path)
-    for existing_model in os.listdir(base_folder):
-        if not os.path.isfile(os.path.join(base_folder, existing_model)):
-            continue
-        existing_model, existing_extension = os.path.splitext(existing_model)
-        if existing_extension != ".zip":
-            continue
-        if SAVING_DELIM in existing_model:
-            *folder, name, i = existing_model.split(SAVING_DELIM)
-            if os.path.join(base_folder, SAVING_DELIM.join(folder)) != saving_path:
-                continue
-            new_path = _make_saving_path(int(i), name, saving_path, False)
-            paths.setdefault(name, []).append(new_path)
-    return paths
-
-
 def load(
     identifier: str = "cflearn",
     saving_folder: Optional[str] = None,
+    *,
+    compress: bool = True,
 ) -> Dict[str, List[Pipeline]]:
     paths = _fetch_saving_paths(identifier, saving_folder)
     pipelines = {
-        k: [Pipeline.load(v, compress=True) for v in v_list]
+        k: [Pipeline.load(v, compress=compress) for v in v_list]
         for k, v_list in paths.items()
     }
     if not pipelines:
@@ -235,41 +275,22 @@ def load(
     return pipelines
 
 
-def _remove(identifier: str = "cflearn", saving_folder: str = None) -> None:
-    for path_list in _fetch_saving_paths(identifier, saving_folder).values():
-        for path in path_list:
-            path = f"{path}.zip"
-            print(f"{LoggingMixin.info_prefix}removing {path}...")
-            os.remove(path)
+def task_loader(saving_folder: str, compress: bool = True) -> Pipeline:
+    return list(load(saving_folder=saving_folder, compress=compress).values())[0][0]
 
 
-def _rmtree(folder: str, patience: float = 10.0) -> None:
-    if not os.path.isdir(folder):
-        return None
-    t = time.time()
-    while True:
-        try:
-            if time.time() - t >= patience:
-                print(f"\n{LoggingMixin.warning_prefix}failed to rmtree: {folder}")
-                break
-            shutil.rmtree(folder)
-            break
-        except:
-            print("", end=".", flush=True)
-            time.sleep(1)
-
-
-def load_task(task: Task) -> Pipeline:
-    return next(iter(load(saving_folder=task.saving_folder).values()))[0]
-
-
-def transform_experiments(experiments: Experiments) -> Dict[str, List[Pipeline]]:
-    return {k: list(map(load_task, v)) for k, v in experiments.tasks.items()}
+def load_experiment_results(results: ExperimentResults) -> Dict[str, List[Pipeline]]:
+    pipelines_dict: Dict[str, Dict[int, Pipeline]] = {}
+    for workplace, workplace_key in zip(results.workplaces, results.workplace_keys):
+        pipeline = task_loader(workplace)
+        model, str_i = workplace_key
+        pipelines_dict.setdefault(model, {})[int(str_i)] = pipeline
+    return {k: [v[i] for i in sorted(v)] for k, v in pipelines_dict.items()}
 
 
 class RepeatResult(NamedTuple):
     data: Optional[DataProtocol]
-    experiments: Optional[Experiments]
+    experiment: Optional[Experiment]
     pipelines: Optional[Dict[str, List[Pipeline]]]
     patterns: Optional[Dict[str, List[ModelPattern]]]
 
@@ -304,23 +325,18 @@ def repeat_with(
     *,
     models: Union[str, List[str]] = "fcnn",
     model_configs: Optional[Dict[str, Dict[str, Any]]] = None,
-    identifiers: Optional[Union[str, List[str]]] = None,
     predict_config: Optional[Dict[str, Any]] = None,
     sequential: Optional[bool] = None,
     num_jobs: int = 1,
     num_repeat: int = 5,
     temp_folder: str = "__tmp__",
     return_patterns: bool = True,
+    compress: bool = True,
     use_tqdm: bool = True,
     **kwargs: Any,
 ) -> RepeatResult:
-
     if isinstance(models, str):
         models = [models]
-    if identifiers is None:
-        identifiers = models.copy()
-    elif isinstance(identifiers, str):
-        identifiers = [identifiers]
 
     kwargs = shallow_copy_dict(kwargs)
     kwargs.setdefault("trigger_logging", False)
@@ -331,33 +347,29 @@ def repeat_with(
     if model_configs is None:
         model_configs = {}
 
+    def fetch_config(model_key: str) -> Dict[str, Any]:
+        local_kwargs = shallow_copy_dict(kwargs)
+        assert model_configs is not None
+        local_model_config = model_configs.setdefault(model_key, {})
+        return update_dict(shallow_copy_dict(local_model_config), local_kwargs)
+
     pipelines_dict: Optional[Dict[str, List[Pipeline]]] = None
     if sequential:
-        experiments = None
+        experiment = None
         kwargs["use_tqdm"] = False
-
         if not return_patterns:
             print(
-                f"{LoggingMixin.warning_prefix}`return_patterns` should be True "
-                "when `sequential` is True, because patterns will always be generated"
+                f"{LoggingMixin.warning_prefix}`return_patterns` should be "
+                "True when `sequential` is True, because patterns "
+                "will always be generated"
             )
             return_patterns = True
-
-        def get(i_: int, model_: str) -> Pipeline:
-            kwargs_ = shallow_copy_dict(kwargs)
-            assert model_configs is not None
-            model_config = model_configs.setdefault(model_, {})
-            kwargs_ = update_dict(shallow_copy_dict(model_config), kwargs_)
-            logging_folder = os.path.join(temp_folder, model_, str(i_))
-            kwargs_.setdefault("logging_folder", logging_folder)
-            m = make(model_, **shallow_copy_dict(kwargs_))
-            return m.fit(x, y, x_cv, y_cv)
-
         pipelines_dict = {}
-        iterator = zip(models, identifiers)
-        if use_tqdm:
-            iterator = tqdm(iterator, total=len(models), position=0)
-        for model, identifier in iterator:
+        if not use_tqdm:
+            iterator = models
+        else:
+            iterator = tqdm(models, total=len(models), position=0)
+        for model in iterator:
             local_pipelines = []
             sub_iterator = range(num_repeat)
             if use_tqdm:
@@ -368,33 +380,42 @@ def repeat_with(
                     leave=False,
                 )
             for i in sub_iterator:
-                local_pipelines.append(get(i, model))
-            pipelines_dict[identifier] = local_pipelines
+                local_config = fetch_config(model)
+                logging_folder = os.path.join(temp_folder, model, str(i))
+                local_config.setdefault("logging_folder", logging_folder)
+                m = make(model, **shallow_copy_dict(local_config))
+                local_pipelines.append(m.fit(x, y, x_cv, y_cv))
+            pipelines_dict[model] = local_pipelines
     else:
         if num_jobs <= 1:
             print(
                 f"{LoggingMixin.warning_prefix}we suggest setting `sequential` "
                 f"to True when `num_jobs` is {num_jobs}"
             )
-
-        experiments = Experiments(temp_folder, overwrite=False)
-        experiments.run(
-            None,
+        # data
+        data_folder = Experiment.dump_data_bundle(
             x,
             y,
             x_cv,
             y_cv,
-            models=models,
-            model_configs=model_configs,
-            identifiers=identifiers,
-            num_repeat=num_repeat,
-            num_jobs=num_jobs,
-            use_tqdm=use_tqdm,
-            temp_folder=temp_folder,
-            **shallow_copy_dict(kwargs),
+            workplace=temp_folder,
         )
+        # experiment
+        experiment = Experiment(num_jobs=num_jobs)
+        for model in models:
+            for _ in range(num_repeat):
+                local_config = fetch_config(model)
+                experiment.add_task(
+                    model=model,
+                    compress=compress,
+                    root_workplace=temp_folder,
+                    config=shallow_copy_dict(local_config),
+                    data_folder=data_folder,
+                )
+        # finalize
+        results = experiment.run_tasks(use_tqdm=use_tqdm)
         if return_patterns:
-            pipelines_dict = transform_experiments(experiments)
+            pipelines_dict = load_experiment_results(results)
 
     patterns = None
     if return_patterns:
@@ -408,18 +429,9 @@ def repeat_with(
 
     data = None
     if patterns is not None:
-        data = patterns[identifiers[0]][0].model.data
+        data = patterns[models[0]][0].model.data
 
-    return RepeatResult(data, experiments, pipelines_dict, patterns)
-
-
-def tasks_to_pipelines(tasks: List[Task]) -> List[Pipeline]:
-    return list(map(load_task, tasks))
-
-
-def tasks_to_patterns(tasks: List[Task], **kwargs: Any) -> List[pattern_type]:
-    pipelines = tasks_to_pipelines(tasks)
-    return [m.to_pattern(**shallow_copy_dict(kwargs)) for m in pipelines]
+    return RepeatResult(data, experiment, pipelines_dict, patterns)
 
 
 def make_toy_model(
@@ -477,21 +489,18 @@ def make_toy_model(
 
 __all__ = [
     "make",
+    "make_from",
+    "finetune",
     "save",
     "load",
     "evaluate",
-    "load_task",
+    "task_loader",
+    "load_experiment_results",
     "repeat_with",
-    "tasks_to_pipelines",
-    "tasks_to_patterns",
-    "transform_experiments",
     "make_toy_model",
     "Task",
-    "Experiments",
+    "Experiment",
     "ModelPattern",
     "EnsemblePattern",
     "RepeatResult",
-    "SAVING_DELIM",
-    "_remove",
-    "_rmtree",
 ]
