@@ -17,6 +17,7 @@ from cftool.misc import LoggingMixin
 from cftool.ml.hpo import HPOBase
 from cftool.ml.utils import pattern_type
 from cftool.ml.utils import scoring_fn_type
+from cftool.ml.utils import Metrics
 from cftool.ml.utils import Estimator
 from cfdata.types import np_float_type
 from cfdata.tabular import task_type_type
@@ -30,10 +31,11 @@ from .ensemble import Ensemble
 from ..types import data_type
 from ..pipeline import Pipeline
 from ..protocol import DataProtocol
+from ..misc._api import SAVING_DELIM
 
 
 class _TunerResult(NamedTuple):
-    identifier: str
+    model: str
     repeat_result: RepeatResult
 
     @property
@@ -41,26 +43,26 @@ class _TunerResult(NamedTuple):
         pipelines = self.repeat_result.pipelines
         if pipelines is None:
             raise ValueError("`pipelines` are not yet generated")
-        return pipelines[self.identifier]
+        return pipelines[self.model]
 
     @property
     def patterns(self) -> List[ModelPattern]:
         patterns = self.repeat_result.patterns
         if patterns is None:
             raise ValueError("`patterns` are not yet generated")
-        return patterns[self.identifier]
+        return patterns[self.model]
 
     @property
-    def weighted_scores(self) -> Dict[str, np.ndarray]:
-        weighted_scores = {}
+    def weighted_metrics(self) -> Dict[str, np.ndarray]:
+        weighted_metrics = {}
         final_results_dict = self.repeat_result.final_results
         if final_results_dict is None:
             raise ValueError("training process is corrupted")
-        final_results = final_results_dict[self.identifier]
+        final_results = final_results_dict[self.model]
         for key in sorted(final_results[0].weighted_scores.keys()):
-            local_scores = [result.weighted_scores[key] for result in final_results]
-            weighted_scores[key] = np.array(local_scores, np_float_type)
-        return weighted_scores
+            local_metrics = [result.weighted_metrics[key] for result in final_results]
+            weighted_metrics[key] = np.array(local_metrics, np_float_type)
+        return weighted_metrics
 
 
 class _Tuner(LoggingMixin):
@@ -93,8 +95,8 @@ class _Tuner(LoggingMixin):
                 read_config["delim"] = delim
             else:
                 print(
-                    f"{LoggingMixin.warning_prefix}delimiter of the given file dataset is not provided, "
-                    "this may cause incorrect parsing"
+                    f"{LoggingMixin.warning_prefix}delimiter of the given "
+                    "file dataset is not provided, this may cause incorrect parsing"
                 )
             if y is not None:
                 read_config["y"] = y
@@ -119,16 +121,28 @@ class _Tuner(LoggingMixin):
     def make_estimators(
         self,
         metrics: Optional[Union[str, List[str]]],
+        repeat_result: Optional[RepeatResult],
     ) -> List[Estimator]:
         if isinstance(metrics, str):
             metrics = [metrics]
         elif metrics is None:
             if self.task_type is None:
                 raise ValueError("either `task_type` or `metrics` should be provided")
-            if parse_task_type(self.task_type) is TaskTypes.CLASSIFICATION:
-                metrics = ["acc", "auc"]
-            else:
-                metrics = ["mae", "mse"]
+            if repeat_result is not None:
+                if repeat_result.pipelines is None:
+                    raise ValueError("pipelines are not provided in `repeat_result`")
+                pipeline = list(repeat_result.pipelines.values())[0][0]
+                metrics = []
+                for metric in sorted(pipeline.trainer.metrics):
+                    if metric in Metrics.sign_dict:
+                        metrics.append(metric)
+                if not metrics:
+                    metrics = None
+            if metrics is None:
+                if parse_task_type(self.task_type) is TaskTypes.CLASSIFICATION:
+                    metrics = ["acc", "auc"]
+                else:
+                    metrics = ["mae", "mse"]
         return list(map(Estimator, metrics))
 
     def make_data(self) -> Tuple[data_type, data_type, data_type, data_type]:
@@ -161,10 +175,10 @@ class _Tuner(LoggingMixin):
         num_parallel: int,
         temp_folder: str,
         *,
+        compress: bool = True,
         cuda: Optional[str] = None,
         sequential: Optional[bool] = None,
     ) -> _TunerResult:
-        identifier = hash_code(str(params))
         params = update_dict(params, shallow_copy_dict(self.base_params))
         params["verbose_level"] = 0
         params["use_tqdm"] = False
@@ -178,13 +192,13 @@ class _Tuner(LoggingMixin):
             num_repeat=num_repeat,
             num_jobs=num_parallel,
             models=model,
-            identifiers=identifier,
+            compress=compress,
             temp_folder=temp_folder,
             predict_config={"contains_labels": True},
             sequential=sequential,
             **params,
         )
-        return _TunerResult(identifier, repeat_result)
+        return _TunerResult(model, repeat_result)
 
     def save(self, export_folder: str) -> "_Tuner":
         Saving.prepare_folder(self, export_folder)
@@ -264,7 +278,8 @@ def tune_with(
 
     if os.path.isdir(temp_folder):
         print(
-            f"{LoggingMixin.warning_prefix}'{temp_folder}' already exists, it will be overwritten"
+            f"{LoggingMixin.warning_prefix}'{temp_folder}' already exists, "
+            "it will be overwritten"
         )
         shutil.rmtree(temp_folder)
 
@@ -301,8 +316,8 @@ def tune_with(
             num_jobs = 0
         if num_jobs > 1:
             print(
-                f"{LoggingMixin.warning_prefix}`num_jobs` is set but hpo is sequential, "
-                "please use `num_parallel` instead"
+                f"{LoggingMixin.warning_prefix}`num_jobs` is set but hpo is "
+                "sequential, please use `num_parallel` instead"
             )
         num_jobs = 0
     if search_config is None:
@@ -319,9 +334,10 @@ def tune_with(
     if num_jobs is not None:
         search_config["num_jobs"] = num_jobs
     search_config.setdefault(
-        "parallel_logging_folder", os.path.join(temp_folder, "__hpo_parallel__")
+        "parallel_logging_folder",
+        os.path.join(temp_folder, "__hpo_parallel__"),
     )
-    estimators = tuner.make_estimators(metrics)
+    estimators = tuner.make_estimators(metrics, None)
     hpo.search(x, y, estimators, x_cv, y_cv, **search_config)
     return HPOResult(hpo, extra_config)
 
@@ -748,6 +764,7 @@ class OptunaPresetParams:
         *,
         tune_lr: bool = True,
         tune_optimizer: bool = True,
+        tune_scheduler: bool = True,
         tune_ema_decay: bool = True,
         tune_clip_norm: bool = True,
         tune_batch_size: bool = True,
@@ -767,6 +784,13 @@ class OptunaPresetParams:
                 "categorical",
             )
             self.base_params["optimizer"] = optimizer_param
+        if tune_scheduler:
+            scheduler_param = OptunaParam(
+                "scheduler",
+                ["step", "plateau", "warmup"],
+                "categorical",
+            )
+            self.base_params["scheduler"] = scheduler_param
         if tune_ema_decay:
             model_config = self.base_params.setdefault("model_config", {})
             assert isinstance(model_config, dict)
@@ -876,6 +900,12 @@ class OptunaPresetParams:
     def _ddr_preset(self) -> optuna_params_type:
         return shallow_copy_dict(self.base_params)
 
+    def _ddr_q_preset(self) -> optuna_params_type:
+        return self._ddr_preset()
+
+    def _ddr_cdf_preset(self) -> optuna_params_type:
+        return self._ddr_preset()
+
     def _rnn_preset(self) -> optuna_params_type:
         return shallow_copy_dict(self.base_params)
 
@@ -885,15 +915,17 @@ class OptunaPresetParams:
 
 class OptunaArgs(NamedTuple):
     cuda: Optional[str]
+    compress: bool
     num_trial: Union[str, int]
-    config: Union[str, Dict[str, Any]]
+    task_config: Union[str, Dict[str, Any]]
     key_mapping: Union[str, OptunaKeyMapping]
 
 
-def optuna_core(args: Union[OptunaArgs, Any]) -> optuna.study.Study:
+def optuna_core(args: OptunaArgs) -> optuna.study.Study:
     cuda = args.cuda
     if cuda == "None":
         cuda = None
+    compress = args.compress
 
     key_mapping_arg = args.key_mapping
     if isinstance(key_mapping_arg, str):
@@ -903,32 +935,40 @@ def optuna_core(args: Union[OptunaArgs, Any]) -> optuna.study.Study:
         key_mapping = key_mapping_arg
     tuner = key_mapping.tuner
 
-    config = args.config
-    if isinstance(config, str):
-        config = Saving.load_dict("config", config)
-    model = config["model"]
-    metrics = config["metrics"]
-    timeout = config["timeout"]
-    num_repeat = config["num_repeat"]
-    num_parallel = config["num_parallel"]
-    estimator_scoring_function = config["estimator_scoring_function"]
-    study_config = config["study_config"]
-    temp_folder = config["temp_folder"]
+    config = args.task_config
+    if isinstance(config, dict):
+        config_dict = config
+    else:
+        config_dict = Saving.load_dict("config", config)
+    model = config_dict["model"]
+    metrics = config_dict["metrics"]
+    timeout = config_dict["timeout"]
+    num_jobs = config_dict["num_jobs"]
+    num_repeat = config_dict["num_repeat"]
+    num_parallel = config_dict["num_parallel"]
+    estimator_scoring_function = config_dict["estimator_scoring_function"]
+    study_config = config_dict["study_config"]
+    temp_folder = config_dict["temp_folder"]
 
     def objective(trial: Trial) -> float:
         temp_folder_ = os.path.join(temp_folder, str(trial.number))
         current_params = key_mapping.pop(trial)
-        current_params["metrics"] = metrics
-        current_params["trial"] = trial
+        if metrics is not None:
+            current_params["metrics"] = metrics
         args_ = model, current_params, num_repeat, num_parallel, temp_folder_
-        result = tuner.train(*args_, sequential=True, cuda=cuda)
+        sequential = None if num_jobs <= 1 else True
+        result = tuner.train(
+            *args_,
+            sequential=sequential,
+            compress=compress,
+            cuda=cuda,
+        )
         final_scores = []
-        weighted_scores = result.weighted_scores
-        estimators = tuner.make_estimators(metrics)
+        weighted_metrics = result.weighted_metrics
+        estimators = tuner.make_estimators(metrics, result.repeat_result)
         for estimator in estimators:
-            metric = estimator.type
             scoring_fn = estimator.scoring_fn(estimator_scoring_function)
-            final_scores.append(scoring_fn(weighted_scores[metric]))
+            final_scores.append(scoring_fn(weighted_metrics[estimator.type]))
         return sum(final_scores) / len(final_scores)
 
     study = optuna.create_study(**study_config)
@@ -952,6 +992,7 @@ def optuna_tune(
     num_trial: int = 50,
     num_repeat: int = 5,
     num_parallel: int = 1,
+    compress: bool = True,
     temp_folder: str = "__tmp__",
     estimator_scoring_function: Union[str, scoring_fn_type] = default_scoring,
     timeout: Optional[float] = None,
@@ -986,17 +1027,21 @@ def optuna_tune(
         storage = None
     else:
         assert isinstance(meta_folder, str)
+        os.makedirs(meta_folder, exist_ok=True)
         storage = os.path.join(meta_folder, "shared.db")
         storage = f"sqlite:///{storage}"
     study_config["storage"] = storage
     study_config["direction"] = "maximize"
     study_config["load_if_exists"] = True
     study_config.setdefault("study_name", f"{model}_optuna")
+    if num_jobs > 1:
+        optuna.create_study(**study_config)
 
     task_config = {
         "model": model,
         "metrics": metrics,
         "timeout": timeout,
+        "num_jobs": num_jobs,
         "num_repeat": num_repeat,
         "num_parallel": num_parallel,
         "estimator_scoring_function": estimator_scoring_function,
@@ -1007,7 +1052,8 @@ def optuna_tune(
     if num_jobs <= 1:
         if isinstance(cuda, int):
             cuda = str(cuda)
-        study = optuna_core(OptunaArgs(cuda, num_trial, task_config, key_mapping))
+        args = OptunaArgs(cuda, compress, num_trial, task_config, key_mapping)
+        study = optuna_core(args)
     else:
         assert isinstance(meta_folder, str)
 
@@ -1018,29 +1064,19 @@ def optuna_tune(
         os.makedirs(task_config_folder, exist_ok=True)
         Saving.save_dict(task_config, "config", task_config_folder)
 
-        def _run(num_trial_: int, cuda: int = None) -> None:
-            python = sys.executable
-            cmd = f"{python} -m {'.'.join(['cflearn', 'dist', 'optuna'])}"
-            cuda_suffix = f"--cuda {cuda}"
-            num_trial_suffix = f"--num_trial {num_trial_}"
-            config_suffix = f"--config {task_config_folder}"
-            key_mapping_suffix = f"--key_mapping {key_mapping_folder}"
-            os.system(
-                f"{cmd} {cuda_suffix} {num_trial_suffix} "
-                f"{config_suffix} {key_mapping_suffix}"
-            )
-
-        parallel = Parallel(
-            num_jobs,
-            use_tqdm=False,
-            use_cuda=torch.cuda.is_available(),
-            logging_folder=meta_folder,
-        )
-
+        experiment = Experiment(num_jobs=num_jobs)
         num_trials = [num_trial // num_jobs] * num_jobs
         num_trials[-1] = num_trial - sum(num_trials[:-1])
-        parallel(_run, num_trials)
-
+        for n in num_trials:
+            experiment.add_task(
+                execute="optuna",
+                root_workplace=temp_folder,
+                num_trial=n,
+                compress=compress,
+                task_config_folder=task_config_folder,
+                key_mapping_folder=key_mapping_folder,
+            )
+        experiment.run_tasks()
         study = optuna.create_study(**study_config)
 
     return OptunaResult(tuner, study, key_mapping)
