@@ -14,6 +14,7 @@ from typing import Union
 from typing import Callable
 from typing import Optional
 from typing import NamedTuple
+from argparse import Namespace
 from cftool.misc import timestamp
 from cftool.misc import update_dict
 from cftool.misc import register_core
@@ -84,8 +85,10 @@ def _parse_config(config: general_config_type) -> Dict[str, Any]:
 
 class Elements(NamedTuple):
     model: str = "fcnn"
+    ds_args: Optional[Namespace] = None
     aggregator: str = "sum"
     aggregator_config: Optional[Dict[str, Any]] = None
+    production: str = "pack"
     data_protocol: str = "tabular"
     loader_protocol: str = "tabular"
     sampler_protocol: str = "tabular"
@@ -114,27 +117,30 @@ class Elements(NamedTuple):
     model_config: Optional[Dict[str, Any]] = None
     loss: str = "auto"
     loss_config: Optional[Dict[str, Any]] = None
+    default_encoding_init_method: Optional[str] = None
     metrics: Union[str, List[str]] = "auto"
     metric_config: Optional[Dict[str, Any]] = None
+    lr: Optional[float] = None
     optimizer: Optional[str] = "adamw"
-    scheduler: Optional[str] = "plateau"
+    scheduler: Optional[str] = "warmup"
     optimizer_config: Optional[Dict[str, Any]] = None
     scheduler_config: Optional[Dict[str, Any]] = None
     optimizers: Optional[Dict[str, Any]] = None
     verbose_level: int = 2
     use_tqdm: bool = True
     trigger_logging: bool = False
+    log_pipeline_to_artifacts: bool = False
     cuda: Optional[Union[int, str]] = None
-    tracker_config: Optional[Dict[str, Any]] = None
+    mlflow_config: Optional[Dict[str, Any]] = None
     extra_config: Optional[Dict[str, Any]] = None
     user_defined_config: Optional[Dict[str, Any]] = None
 
     @staticmethod
     def affected_mappings() -> Dict[str, Set[str]]:
         return {
-            "metrics": {"metric_config"},
+            "metrics": {"metric_config*"},
             "use_simplify_data": {"simplify"},
-            "ts_config": {"time_series_config"},
+            "ts_config": {"time_series_config*"},
             "fixed_epoch": {"min_epoch", "num_epoch", "max_epoch"},
         }
 
@@ -173,7 +179,7 @@ class Elements(NamedTuple):
             kwargs["logging_file"] = logging_file
         # trainer
         trainer_config = kwargs.setdefault("trainer_config", {})
-        trainer_config["use_amp"] = kwargs.pop("use_amp")
+        trainer_config.setdefault("use_amp", kwargs.pop("use_amp"))
         min_epoch = kwargs.pop("min_epoch")
         num_epoch = kwargs.pop("num_epoch")
         max_epoch = kwargs.pop("max_epoch")
@@ -213,11 +219,11 @@ class Elements(NamedTuple):
             if min_epoch > num_epoch:
                 raise ValueError("`min_epoch` should not be greater than `num_epoch`")
             max_epoch = max(default_max, num_epoch)
-        trainer_config["min_epoch"] = min_epoch
-        trainer_config["num_epoch"] = num_epoch
-        trainer_config["max_epoch"] = max_epoch
-        trainer_config["max_snapshot_file"] = kwargs.pop("max_snapshot_file")
-        trainer_config["clip_norm"] = kwargs.pop("clip_norm")
+        trainer_config.setdefault("min_epoch", min_epoch)
+        trainer_config.setdefault("num_epoch", num_epoch)
+        trainer_config.setdefault("max_epoch", max_epoch)
+        trainer_config.setdefault("max_snapshot_file", kwargs.pop("max_snapshot_file"))
+        trainer_config.setdefault("clip_norm", kwargs.pop("clip_norm"))
         # model
         model_config = self.model_config or {}
         model_config["aggregator"] = kwargs.pop("aggregator")
@@ -225,13 +231,19 @@ class Elements(NamedTuple):
         model_config["ema_decay"] = kwargs.pop("ema_decay")
         model_config["loss"] = kwargs.pop("loss")
         model_config["loss_config"] = kwargs.pop("loss_config") or {}
+        default_encoding_init_method = kwargs.pop("default_encoding_init_method")
+        if default_encoding_init_method is not None:
+            de_cfg = model_config.setdefault("default_encoding_configs", {})
+            de_cfg["init_method"] = default_encoding_init_method
         kwargs["model_config"] = model_config
         # metrics
         metric_config = kwargs.pop("metric_config") or {}
         if self.metric_config is None and self.metrics is not None:
             metric_config["types"] = kwargs.pop("metrics")
-        trainer_config["metric_config"] = metric_config
+        metric_config.setdefault("decay", 0.1)
+        trainer_config.setdefault("metric_config", metric_config)
         # optimizers
+        lr = kwargs.pop("lr")
         optimizer = kwargs.pop("optimizer")
         scheduler = kwargs.pop("scheduler")
         optimizers = kwargs.pop("optimizers")
@@ -263,24 +275,18 @@ class Elements(NamedTuple):
                     "so `scheduler_config` will be ignored"
                 )
         else:
-            preset_optimizer = {}
-            if optimizer is not None:
-                optimizer_config.setdefault("lr", 1e-3)
-                preset_optimizer = {
+            if lr is not None:
+                optimizer_config["lr"] = lr
+            optimizers = {
+                "all": {
                     "optimizer": optimizer,
+                    "scheduler": scheduler,
                     "optimizer_config": optimizer_config,
+                    "scheduler_config": scheduler_config,
                 }
-            if scheduler is not None:
-                preset_optimizer.update(
-                    {
-                        "scheduler": scheduler,
-                        "scheduler_config": scheduler_config,
-                    }
-                )
-            if preset_optimizer:
-                optimizers = {"all": preset_optimizer}
+            }
         if optimizers is not None:
-            trainer_config["optimizers"] = optimizers
+            trainer_config.setdefault("optimizers", optimizers)
         # inject user defined configs
         update_dict(self.user_defined_config, kwargs)
         return kwargs
@@ -323,9 +329,6 @@ class Elements(NamedTuple):
         trainer_config["use_amp"] = use_amp and amp is not None
         default_checkpoint_folder = os.path.join(log_folder, "checkpoints")
         trainer_config.setdefault("checkpoint_folder", default_checkpoint_folder)
-        # trainer -> metrics
-        metric_config = trainer_config.setdefault("metric_config", {})
-        metric_config.setdefault("decay", 0.1)
         # model general
         model_config = kwargs.setdefault("model_config", {})
         encoding_methods = model_config.get("encoding_methods", {})
@@ -341,7 +344,7 @@ class Elements(NamedTuple):
         kwargs.setdefault("trial", None)
         kwargs.setdefault("use_tqdm", True)
         kwargs.setdefault("verbose_level", 2)
-        kwargs.setdefault("tracker_config", None)
+        kwargs.setdefault("mlflow_config", None)
         kwargs.setdefault("trigger_logging", False)
         kwargs.setdefault("use_timing_context", True)
         # convert to `Elements`
@@ -361,11 +364,21 @@ class Environment:
         self,
         config: Dict[str, Any],
         user_defined_config: Optional[Dict[str, Any]] = None,
+        set_device: bool = True,
     ):
         self.config = config
-        if user_defined_config is None:
-            user_defined_config = config
-        self.user_defined_config = user_defined_config
+        self.user_defined_config = user_defined_config or {}
+        # deep speed
+        self.is_rank_0 = True
+        if self.deepspeed:
+            self.is_rank_0 = self.local_rank is None or self.local_rank == 0
+            logging_folder = config.pop("logging_folder")
+            current_timestamp = timestamp(ensure_different=True)
+            config["logging_folder"] = os.path.join(logging_folder, current_timestamp)
+            if set_device:
+                if self.local_rank is None:
+                    self.ds_args.local_rank = 0
+                torch.cuda.set_device(self.local_rank)
 
     def __getattr__(self, item: str) -> Any:
         return self.config[item]
@@ -385,17 +398,30 @@ class Environment:
                 current_v = current.get(k)
                 if current_v is None:
                     current[k] = new_default_v
-                elif k not in user_affected:
+                else:
+                    if f"{k}*" in user_affected:
+                        continue
                     if not isinstance(new_default_v, dict):
-                        current[k] = new_default_v
+                        if k not in user_affected:
+                            current[k] = new_default_v
                     else:
                         _core(current_v, new_default_v)
 
         _inject_affected(self.user_defined_config)
         _core(self.config, new_default_config)
 
+    @property
+    def deepspeed(self) -> bool:
+        return self.ds_args is not None
+
+    @property
+    def local_rank(self) -> int:
+        return self.ds_args.local_rank
+
     @property
     def device(self) -> torch.device:
+        if self.deepspeed:
+            return torch.device("cuda", self.local_rank)
         cuda = self.cuda
         if cuda == "cpu":
             return torch.device("cpu")
@@ -416,8 +442,12 @@ class Environment:
         return self.config["model_config"]
 
     @classmethod
-    def from_elements(cls, elements: Elements) -> "Environment":
-        return cls(elements.to_config(), elements.user_defined_config)
+    def from_elements(
+        cls,
+        elements: Elements,
+        set_device: bool = True,
+    ) -> "Environment":
+        return cls(elements.to_config(), elements.user_defined_config, set_device)
 
 
 __all__ = ["configs_dict", "Configs", "Elements", "Environment"]
