@@ -1,6 +1,9 @@
+import os
 import math
 import torch
+import inspect
 import logging
+import platform
 
 import numpy as np
 import torch.nn as nn
@@ -8,6 +11,7 @@ import torch.nn as nn
 from typing import *
 from abc import abstractmethod
 from abc import ABCMeta
+from argparse import Namespace
 from functools import partial
 from cftool.misc import context_error_handler
 from cftool.misc import LoggingMixin
@@ -112,7 +116,85 @@ def get_gradient(
     return grads
 
 
-class Initializer(LoggingMixin):
+def scheduler_requires_metric(scheduler: Any) -> bool:
+    signature = inspect.signature(scheduler.step)
+    for name, param in signature.parameters.items():
+        if param.kind is inspect.Parameter.POSITIONAL_OR_KEYWORD:
+            if name == "metrics":
+                return True
+    return False
+
+
+def parse_uri(path: str) -> str:
+    delim = "/" if platform.system() == "Windows" else ""
+    return f"file://{delim}{path}"
+
+
+def to_relative(abs_folder: str, root_abs_folder: str) -> str:
+    common_prefix = os.path.commonpath([root_abs_folder, abs_folder])
+    return os.path.relpath(abs_folder, common_prefix)
+
+
+def parse_args(args: Any) -> Namespace:
+    return Namespace(**{k: None if not v else v for k, v in args.__dict__.items()})
+
+
+def parse_path(path: Optional[str], root_dir: str) -> Optional[str]:
+    if path is None:
+        return None
+    return os.path.abspath(os.path.join(root_dir, path))
+
+
+class LoggingMixinWithRank(LoggingMixin):
+    is_rank_0: bool = True
+
+    def set_rank_0(self, value: bool) -> None:
+        self.is_rank_0 = value
+        for v in self.__dict__.values():
+            if isinstance(v, LoggingMixinWithRank):
+                v.set_rank_0(value)
+
+    def _init_logging(
+        self,
+        verbose_level: Optional[int] = 2,
+        trigger: bool = True,
+    ) -> None:
+        if not self.is_rank_0:
+            return None
+        super()._init_logging(verbose_level, trigger)
+
+    def log_msg(
+        self,
+        body: str,
+        prefix: str = "",
+        verbose_level: Optional[int] = 1,
+        msg_level: int = logging.INFO,
+        frame: Any = None,
+    ) -> None:
+        if not self.is_rank_0:
+            return None
+        super().log_msg(body, prefix, verbose_level, msg_level, frame)
+
+    def log_block_msg(
+        self,
+        body: str,
+        prefix: str = "",
+        title: str = "",
+        verbose_level: Optional[int] = 1,
+        msg_level: int = logging.INFO,
+        frame: Any = None,
+    ) -> None:
+        if not self.is_rank_0:
+            return None
+        super().log_block_msg(body, prefix, title, verbose_level, msg_level, frame)
+
+    def log_timing(self) -> None:
+        if not self.is_rank_0:
+            return None
+        return super().log_timing()
+
+
+class Initializer(LoggingMixinWithRank):
     """
     Initializer for neural network weights
 
@@ -556,7 +638,13 @@ __all__ = [
     "collate_tensor_dicts",
     "switch_requires_grad",
     "get_gradient",
+    "scheduler_requires_metric",
+    "parse_uri",
+    "to_relative",
+    "parse_args",
+    "parse_path",
     "Lambda",
+    "LoggingMixinWithRank",
     "Initializer",
     "Activations",
     "mode_context",
