@@ -1,6 +1,10 @@
 from typing import *
 from torch.optim import Optimizer
-from torch.optim.lr_scheduler import _LRScheduler, ReduceLROnPlateau
+from torch.optim.lr_scheduler import _LRScheduler
+from torch.optim.lr_scheduler import StepLR
+from torch.optim.lr_scheduler import ReduceLROnPlateau
+
+from ..misc.toolkit import scheduler_requires_metric
 
 scheduler_dict = {}
 
@@ -14,7 +18,35 @@ def register_scheduler(name: str) -> Callable[[Type], Type]:
     return _register
 
 
-register_scheduler("plateau")(ReduceLROnPlateau)
+@register_scheduler("step")
+class StepLRWithFloor(StepLR):
+    def __init__(
+        self,
+        optimizer: Optimizer,
+        step_size: int,
+        gamma: float = 0.1,
+        last_epoch: int = -1,
+        lr_floor: float = 1.0e-8,
+    ):
+        self.lr_floor = lr_floor
+        super().__init__(optimizer, step_size, gamma, last_epoch)
+
+    def get_lr(self) -> List[float]:  # type: ignore
+        lrs = super().get_lr()
+        return [max(lr, self.lr_floor) for lr in lrs]  # type: ignore
+
+    def _get_closed_form_lr(self) -> List[float]:  # type: ignore
+        lrs = super()._get_closed_form_lr()  # type: ignore
+        return [max(lr, self.lr_floor) for lr in lrs]
+
+
+@register_scheduler("plateau")
+class ReduceLROnPlateauWithGet(ReduceLROnPlateau):
+    def get_lr(self) -> List[float]:
+        return [group["lr"] for group in self.optimizer.param_groups]  # type: ignore
+
+    def get_last_lr(self) -> List[float]:
+        return self.get_lr()
 
 
 @register_scheduler("warmup")
@@ -35,42 +67,20 @@ class WarmupScheduler(_LRScheduler):
         self.scheduler_afterwards = scheduler_afterwards_base(
             optimizer, **scheduler_afterwards_config
         )
+        self.requires_metric = scheduler_requires_metric(self.scheduler_afterwards)
         super().__init__(optimizer)
 
     @property
     def lr_warmup_func(self) -> Callable[[float], float]:
-        multiplier = (self.multiplier - 1.0) * self.last_epoch / self.warmup_step + 1.0
+        multiplier = (self.multiplier - 1.0) * self.last_epoch / self.warmup_step + 1.0  # type: ignore
         return lambda lr: lr * multiplier
 
     @property
     def lr_multiplier_func(self) -> Callable[[float], float]:
         return lambda lr: lr * self.multiplier
 
-    @property
-    def reduce_on_plateau_afterwards(self) -> bool:
-        return isinstance(self, ReduceLROnPlateau)
-
-    def _step_reduce_on_plateau(
-        self,
-        metrics: float,
-        epoch: Optional[int] = None,
-    ) -> None:
-        if epoch is None:
-            epoch = self.last_epoch + 1
-        self.last_epoch: int = epoch
-        if self.last_epoch <= self.warmup_step:
-            for param_group, lr in zip(
-                self.optimizer.param_groups,  # type: ignore
-                map(self.lr_warmup_func, self.base_lrs),  # type: ignore
-            ):
-                param_group["lr"] = lr
-        else:
-            if epoch is not None:
-                epoch -= self.warmup_step
-            self.scheduler_afterwards.step(metrics, None)  # type: ignore
-
-    def get_lr(self) -> Union[float, List[float]]:  # type: ignore
-        if self.last_epoch > self.warmup_step:
+    def get_lr(self) -> List[float]:  # type: ignore
+        if self.last_epoch > self.warmup_step:  # type: ignore
             if self.scheduler_afterwards is not None:
                 if not self.finished_warmup:
                     self.finished_warmup = True
@@ -78,24 +88,23 @@ class WarmupScheduler(_LRScheduler):
                         map(self.lr_multiplier_func, self.base_lrs)  # type: ignore
                     )
                     self.scheduler_afterwards.base_lrs = base_lrs  # type: ignore
-                return self.scheduler_afterwards.get_lr()
+                return self.scheduler_afterwards.get_lr()  # type: ignore
             return list(map(self.lr_multiplier_func, self.base_lrs))
         return list(map(self.lr_warmup_func, self.base_lrs))  # type: ignore
 
-    def step(
-        self,
-        epoch: Optional[int] = None,
-        metrics: Optional[float] = None,
-    ) -> None:
-        if self.reduce_on_plateau_afterwards:
-            assert metrics is not None
-            self._step_reduce_on_plateau(metrics, epoch)
+    def get_last_lr(self) -> List[float]:
+        if not self.finished_warmup:
+            return super().get_last_lr()  # type: ignore
+        return self.scheduler_afterwards.get_last_lr()  # type: ignore
+
+    def step(self, metrics: Optional[float] = None) -> None:
+        if not self.finished_warmup or self.scheduler_afterwards is None:
+            return super().step()
+        if not self.requires_metric:
+            self.scheduler_afterwards.step()
         else:
-            if not self.finished_warmup or self.scheduler_afterwards is None:
-                return super().step(epoch)
-            if epoch is not None:
-                epoch -= self.warmup_step
-            self.scheduler_afterwards.step(epoch)
+            assert metrics is not None
+            self.scheduler_afterwards.step(metrics)  # type: ignore
 
 
 __all__ = ["scheduler_dict", "register_scheduler"]
