@@ -1,5 +1,7 @@
 import os
+import json
 import torch
+import shutil
 
 import numpy as np
 import matplotlib.pyplot as plt
@@ -20,16 +22,20 @@ try:
 except:
     amp = None
 
-from .data import PrefetchLoader
 from .types import data_type
 from .configs import Elements
 from .configs import Environment
 from .trainer import Trainer
+from .trainer import IntermediateResults
 from .protocol import DataProtocol
+from .protocol import PrefetchLoader
 from .protocol import DataLoaderProtocol
 from .inference import Inference
 from .inference import PreProcessor
+from .misc._api import _fetch_saving_paths
 from .misc.toolkit import to_2d
+from .misc.toolkit import to_relative
+from .misc.toolkit import LoggingMixinWithRank
 from .misc.time_series import TSLabelCollator
 from .models.base import model_dict
 from .models.base import ModelBase
@@ -38,7 +44,7 @@ from .models.base import ModelBase
 key_type = Tuple[Union[str, Optional[str]], ...]
 
 
-class Pipeline(LoggingMixin):
+class Pipeline(LoggingMixinWithRank):
     def __init__(self, environment: Environment):
         # typing
         self.tr_data: DataProtocol
@@ -58,7 +64,6 @@ class Pipeline(LoggingMixin):
         self.data_config["use_timing_context"] = self.timing
         self.data_config["default_categorical_process"] = "identical"
         self.sampler_config = self.config.setdefault("sampler_config", {})
-        self._init_logging(environment.verbose_level, environment.trigger_logging)
 
     def __getattr__(self, item: str) -> Any:
         return self.environment.config.get(item)
@@ -151,6 +156,17 @@ class Pipeline(LoggingMixin):
         self.tr_loader_copy.sampler.shuffle = False
 
     def _prepare_modules(self, *, is_loading: bool = False) -> None:
+        # logging
+        if not is_loading:
+            if os.path.isdir(self.logging_folder):
+                if os.listdir(self.logging_folder):
+                    print(
+                        f"{self.warning_prefix}'{self.logging_folder}' already exists, "
+                        "it will be cleared up to store our logging"
+                    )
+                shutil.rmtree(self.logging_folder)
+            os.makedirs(self.logging_folder)
+        self._init_logging(self.verbose_level, self.trigger_logging)
         # model
         with timing_context(self, "init model", enable=self.timing):
             self.model = model_dict[self.model_type](
@@ -181,7 +197,7 @@ class Pipeline(LoggingMixin):
             )
         # to device
         with timing_context(self, "init device", enable=self.timing):
-            self.trainer.model.to(self.device)
+            self.model.to(self.device)
 
     def _before_loop(
         self,
@@ -225,17 +241,45 @@ class Pipeline(LoggingMixin):
                 if sample_weights is not None:
                     self.tr_weights = sample_weights[split.remained_indices]
                     self.cv_weights = sample_weights[split.split_indices]
+        # deep speed
+        self.set_rank_0(self.environment.is_rank_0)
+        # data
         self._init_data()
         # modules
         self._prepare_modules()
+        # deep speed
+        self.set_rank_0(self.is_rank_0)
+
+    def _handle_pretrain(
+        self,
+        strict: bool,
+        folder: Optional[str],
+        identifier: Optional[str],
+        state_dict_callback: Optional[Callable[[Dict[str, Any]], None]],
+    ) -> None:
+        if identifier is None:
+            return None
+        paths_dict = _fetch_saving_paths(identifier, folder)
+        all_paths: List[str] = sum(paths_dict.values(), [])
+        if len(all_paths) > 1:
+            raise ValueError("more than 1 model is detected")
+        path = all_paths[0]
+        folder = folder or "./"
+        compress = os.path.isfile(f"{path}.zip")
+        with lock_manager(folder, [path]):
+            with Saving.compress_loader(path, compress):
+                self.trainer.restore_checkpoint(path, strict, state_dict_callback)
 
     def _loop(self) -> None:
         # dump information
         logging_folder = self.logging_folder
         os.makedirs(logging_folder, exist_ok=True)
-        Saving.save_dict(self.config, "config", logging_folder)
-        with open(os.path.join(logging_folder, "model.txt"), "w") as f:
-            f.write(str(self.model))
+        if self.is_rank_0:
+            if self.environment.deepspeed:
+                logging_folder = os.path.join(logging_folder, os.pardir)
+            Saving.save_dict(self.config, "config", logging_folder)
+            with open(os.path.join(logging_folder, "model.txt"), "w") as f:
+                f.write(str(self.model))
         # training loop
         self.trainer.fit(
             self.tr_loader,
@@ -244,7 +288,6 @@ class Pipeline(LoggingMixin):
             self.tr_weights,
             self.cv_weights,
         )
-        # logging
         self.log_timing()
 
     @staticmethod
@@ -463,10 +506,62 @@ class Pipeline(LoggingMixin):
         x_cv: data_type = None,
         y_cv: data_type = None,
         *,
+        pretrain_strict: bool = False,
+        pretrain_folder: Optional[str] = None,
+        pretrain_identifier: Optional[str] = None,
+        state_dict_callback: Optional[Callable[[Dict[str, Any]], None]] = None,
         sample_weights: Optional[np.ndarray] = None,
     ) -> "Pipeline":
         self._before_loop(x, y, x_cv, y_cv, sample_weights)
+        self._handle_pretrain(
+            pretrain_strict,
+            pretrain_folder,
+            pretrain_identifier,
+            state_dict_callback,
+        )
         self._loop()
+        # finalize mlflow
+        run_id = self.trainer.run_id
+        mlflow_client = self.trainer.mlflow_client
+        if mlflow_client is not None:
+            # log model
+            import mlflow
+            import cflearn
+
+            cwd = os.getcwd()
+            root_folder = os.path.join(os.path.dirname(__file__), os.pardir)
+            conda_env = os.path.join(os.path.abspath(root_folder), "conda.yml")
+            if self.production == "pack":
+                pack_folder = os.path.join(self.logging_folder, "__packed__")
+                pack_folder = to_relative(os.path.abspath(pack_folder), cwd)
+                cflearn.Pack.pack(self, pack_folder, compress=False, verbose=False)
+                mlflow.pyfunc.save_model(
+                    os.path.join(self.logging_folder, "__pyfunc__"),
+                    python_model=cflearn.PackModel(),
+                    artifacts={"pack_folder": pack_folder},
+                    conda_env=conda_env,
+                )
+                cflearn._rmtree(pack_folder)
+            elif self.production == "pipeline":
+                export_folder = os.path.join(self.logging_folder, "pipeline")
+                export_folder = to_relative(os.path.abspath(export_folder), cwd)
+                self.save(export_folder, compress=False)
+                mlflow.pyfunc.save_model(
+                    os.path.join(self.logging_folder, "__pyfunc__"),
+                    python_model=cflearn.PipelineModel(),
+                    artifacts={"export_folder": export_folder},
+                    conda_env=conda_env,
+                )
+            else:
+                msg = f"unrecognized production type '{self.production}' found"
+                raise NotImplementedError(msg)
+            # log artifacts
+            if self.environment.log_pipeline_to_artifacts:
+                if self.production != "pipeline":
+                    self.save(os.path.join(self.logging_folder, "pipeline"))
+            mlflow_client.log_artifacts(run_id, self.logging_folder)
+            # terminate
+            mlflow_client.set_terminated(run_id)
         return self
 
     def predict(
@@ -548,6 +643,7 @@ class Pipeline(LoggingMixin):
     train_indices_file = "train_indices.npy"
     valid_indices_file = "valid_indices.npy"
     sample_weights_file = "sample_weights.npy"
+    final_results_file = "final_results.json"
 
     @classmethod
     def make(cls, config: Dict[str, Any]) -> "Pipeline":
@@ -558,7 +654,6 @@ class Pipeline(LoggingMixin):
         export_folder: Optional[str] = None,
         *,
         compress: bool = True,
-        retain_data: bool = True,
         remove_original: bool = True,
     ) -> "Pipeline":
         if export_folder is None:
@@ -566,7 +661,6 @@ class Pipeline(LoggingMixin):
         abs_folder = os.path.abspath(export_folder)
         base_folder = os.path.dirname(abs_folder)
         with lock_manager(base_folder, [export_folder]):
-            Saving.prepare_folder(self, export_folder)
             data_folder = os.path.join(export_folder, self.data_folder)
             os.makedirs(data_folder, exist_ok=True)
             if self.sample_weights is not None:
@@ -576,23 +670,11 @@ class Pipeline(LoggingMixin):
                 assert self.cv_data is not None
                 train_data_folder = os.path.join(data_folder, self.train_folder)
                 valid_data_folder = os.path.join(data_folder, self.valid_folder)
-                self.tr_data.save(
-                    train_data_folder,
-                    retain_data=retain_data,
-                    compress=False,
-                )
-                self.cv_data.save(
-                    valid_data_folder,
-                    retain_data=retain_data,
-                    compress=False,
-                )
+                self.tr_data.save(train_data_folder, compress=False)
+                self.cv_data.save(valid_data_folder, compress=False)
             else:
                 original_data_folder = os.path.join(data_folder, self.original_folder)
-                self._original_data.save(
-                    original_data_folder,
-                    retain_data=retain_data,
-                    compress=False,
-                )
+                self._original_data.save(original_data_folder, compress=False)
                 if self.tr_split_indices is not None:
                     tr_file = os.path.join(data_folder, self.train_indices_file)
                     np.save(tr_file, self.tr_split_indices)
@@ -602,6 +684,8 @@ class Pipeline(LoggingMixin):
             final_results = self.trainer.final_results
             if final_results is None:
                 raise ValueError("`final_results` are not generated yet")
+            with open(os.path.join(export_folder, self.final_results_file), "w") as f:
+                json.dump(final_results, f)
             score = final_results.final_score
             self.trainer.save_checkpoint(score, export_folder)
             if self.inference is None:
@@ -613,19 +697,12 @@ class Pipeline(LoggingMixin):
         return self
 
     @classmethod
-    def load(
-        cls,
-        export_folder: str,
-        *,
-        compress: bool = True,
-        verbose_level: int = 0,
-        cuda: Optional[Union[int, str]] = None,
-    ) -> "Pipeline":
+    def load(cls, export_folder: str, *, compress: bool = True) -> "Pipeline":
         base_folder = os.path.dirname(os.path.abspath(export_folder))
         with lock_manager(base_folder, [export_folder]):
             with Saving.compress_loader(export_folder, compress):
                 config = Saving.load_dict("config", export_folder)
-                config.update({"verbose_level": verbose_level, "cuda": cuda})
+                config.update({"verbose_level": 0})
                 pipeline = cls.make(config)
                 data_folder = os.path.join(export_folder, cls.data_folder)
                 # sample weights
@@ -681,6 +758,7 @@ class Pipeline(LoggingMixin):
                 pipeline._init_data()
                 pipeline._prepare_modules(is_loading=True)
                 trainer = pipeline.trainer
+                trainer.state.inject_loader(pipeline.tr_loader)
                 trainer.tr_loader = PrefetchLoader(pipeline.tr_loader, pipeline.device)
                 cv_loader = pipeline.cv_loader
                 if cv_loader is None:
@@ -689,6 +767,9 @@ class Pipeline(LoggingMixin):
                     trainer.cv_loader = PrefetchLoader(cv_loader, pipeline.device)
                 trainer.restore_checkpoint(export_folder)
                 trainer._init_metrics()
+                final_results_path = os.path.join(export_folder, cls.final_results_file)
+                with open(final_results_path, "r") as f:
+                    trainer.final_results = IntermediateResults(*json.load(f))
         return pipeline
 
 
